@@ -5,8 +5,8 @@ import time
 from copy import copy
 from datetime import datetime, timedelta
 
-from sqlalchemy import (Column, Integer, String, Unicode, DateTime, Boolean,
-                        desc, select, update, delete, ForeignKey, Index, func, and_, or_)
+from sqlalchemy import (Column, Integer, String, Unicode, DateTime, Date, Boolean,
+                        desc, select, update, delete, ForeignKey, Index, func, and_, not_)
 from sqlalchemy.orm import relation, join
 from sqlalchemy.ext.hybrid import Comparator, hybrid_property
 from sqlalchemy.exc import OperationalError
@@ -22,10 +22,10 @@ from flexget.utils.sqlalchemy_utils import (table_columns, table_exists, drop_ta
 from flexget.utils.tools import merge_dict_from_to, parse_timedelta
 from flexget.utils.database import quality_property
 from flexget.manager import Session
-from flexget.plugin import (register_plugin, register_parser_option, get_plugin_by_name, get_plugin_keywords,
-                            DependencyError, priority)
+from flexget.plugin import (register_plugin, register_parser_option, get_plugin_by_name, DependencyError, priority,
+                            PluginError)
 
-SCHEMA_VER = 9
+SCHEMA_VER = 10
 
 log = logging.getLogger('series')
 Base = db_schema.versioned_base('series', SCHEMA_VER)
@@ -132,6 +132,9 @@ def upgrade(ver, session):
                 session.execute(delete(series_table, series_table.c.id.in_(ids[1:])))
             session.execute(update(series_table, series_table.c.id == ids[0], {'name_lower': series}))
         ver = 9
+    if ver == 9:
+        table_add_column('series', 'begin_episode_id', Integer, session)
+        ver = 10
 
     return ver
 
@@ -199,7 +202,11 @@ class Series(Base):
     _name = Column('name', Unicode)
     _name_normalized = Column('name_lower', Unicode, index=True, unique=True)
     identified_by = Column(String)
-    episodes = relation('Episode', backref='series', cascade='all, delete, delete-orphan')
+    begin_episode_id = Column(Integer, ForeignKey('series_episodes.id', name='begin_episode_id', use_alter=True))
+    begin = relation('Episode', uselist=False, primaryjoin="Series.begin_episode_id == Episode.id", foreign_keys=[begin_episode_id])
+    episodes = relation('Episode', backref='series', cascade='all, delete, delete-orphan',
+                        primaryjoin='Series.id == Episode.series_id')
+    in_tasks = relation('SeriesTask', backref='series', cascade='all, delete, delete-orphan')
 
     # Make a special property that does indexed case insensitive lookups on name, but stores/returns specified case
     def name_getter(self):
@@ -238,6 +245,8 @@ class Episode(Base):
 
     @hybrid_property
     def first_seen(self):
+        if not self.releases:
+            return None
         return min(release.first_seen for release in self.releases)
 
     @first_seen.expression
@@ -280,6 +289,26 @@ class Episode(Base):
     def __repr__(self):
         return unicode(self).encode('ascii', 'replace')
 
+    def __eq__(self, other):
+        if not isinstance(other, Episode):
+            return NotImplemented
+        if self.identified_by != other.identified_by:
+            return NotImplemented
+        return self.identifier == other.identifier
+
+    def __lt__(self, other):
+        if not isinstance(other, Episode):
+            return NotImplemented
+        if self.identified_by != other.identified_by:
+            return NotImplemented
+        if self.identified_by in ['ep', 'sequence']:
+            return self.season < other.season or (self.season == other.season and self.number < other.number)
+        if self.identified_by == 'date':
+            return self.identifier < other.identifier
+        # Can't compare id type identifiers
+        return NotImplemented
+
+
 Index('episode_series_identifier', Episode.series_id, Episode.identifier)
 
 
@@ -314,16 +343,27 @@ class Release(Base):
         return unicode(self).encode('ascii', 'replace')
 
 
+class SeriesTask(Base):
+    __tablename__ = 'series_tasks'
+
+    id = Column(Integer, primary_key=True)
+    series_id = Column(Integer, ForeignKey('series.id'), nullable=False)
+    name = Column(Unicode, index=True)
+
+    def __init__(self, name):
+        self.name = name
+
+
 class SeriesDatabase(object):
 
     """Provides API to series database"""
 
-    def get_latest_info(self, series):
+    def get_latest_episode(self, series):
         """Return latest known identifier in dict (season, episode, name) for series name"""
         session = Session.object_session(series)
-        episode = session.query(Episode).select_from(join(Episode, Series)).\
-            filter(Episode.season != None).\
+        episode = session.query(Episode).join(Episode.series).\
             filter(Series.id == series.id).\
+            filter(Episode.season != None).\
             order_by(desc(Episode.season)).\
             order_by(desc(Episode.number)).first()
         if not episode:
@@ -331,7 +371,7 @@ class SeriesDatabase(object):
             return False
         # log.trace('get_latest_info, series: %s season: %s episode: %s' % \
         #    (name, episode.season, episode.number))
-        return {'season': episode.season, 'episode': episode.number, 'name': series.name}
+        return episode
 
     def auto_identified_by(self, series):
         """
@@ -342,7 +382,7 @@ class SeriesDatabase(object):
         """
 
         session = Session.object_session(series)
-        type_totals = dict(session.query(Episode.identified_by, func.count(Episode.identified_by)).join(Series).
+        type_totals = dict(session.query(Episode.identified_by, func.count(Episode.identified_by)).join(Episode.series).
                            filter(Series.id == series.id).group_by(Episode.identified_by).all())
         # Remove None and specials from the dict,
         # we are only considering episodes that we know the type of (parsed with new parser)
@@ -377,7 +417,7 @@ class SeriesDatabase(object):
         :return: Instance of Episode or None if not found.
         """
         session = Session.object_session(series)
-        downloaded = session.query(Episode).join(Release, Series).\
+        downloaded = session.query(Episode).join(Episode.releases, Episode.series).\
             filter(Series.id == series.id).\
             filter(Release.downloaded == True)
         if series.identified_by in ['ep', 'sequence']:
@@ -400,7 +440,7 @@ class SeriesDatabase(object):
         """
         session = Session.object_session(since_ep)
         series = since_ep.series
-        series_eps = session.query(Episode).select_from(join(Episode, Series)).\
+        series_eps = session.query(Episode).join(Episode.series).\
             filter(Series.id == series.id)
         if series.identified_by == 'ep':
             if since_ep.season is None or since_ep.number is None:
@@ -460,7 +500,7 @@ class SeriesDatabase(object):
                 series.episodes.append(episode)  # pylint:disable=E1103
                 log.debug('-> added %s' % episode)
 
-            # if release does not exists in episodes, add new
+            # if release does not exists in episode, add new
             #
             # NOTE:
             #
@@ -484,12 +524,50 @@ class SeriesDatabase(object):
         return releases
 
 
+def set_series_begin(series, ep_id):
+    # If identified_by is not explicitly specified, auto-detect it based on begin identifier
+    # TODO: use some method of series parser to do the identifier parsing
+    session = Session.object_session(series)
+    if isinstance(ep_id, int):
+        identified_by = 'sequence'
+    elif re.match(r'(?i)^S\d{1,2}E\d{1,2}$', ep_id):
+        identified_by = 'ep'
+        ep_id = ep_id.upper()
+    elif re.match(r'\d{4}-\d{2}-\d{2}', ep_id):
+        identified_by = 'date'
+    else:
+        raise ValueError('`%s` is not a valid episode identifier' % ep_id)
+    if series.identified_by not in ['auto', '', None]:
+        if identified_by != series.identified_by:
+            raise ValueError('`begin` value `%s` does not match identifier type for identified_by `%s`' %
+                              (ep_id, series.identified_by))
+    series.identified_by = identified_by
+    episode = (session.query(Episode).filter(Episode.series_id == series.id).
+               filter(Episode.identified_by == series.identified_by).
+               filter(Episode.identifier == str(ep_id)).first())
+    if not episode:
+        # TODO: Don't duplicate code from self.store method
+        episode = Episode()
+        episode.identifier = ep_id
+        episode.identified_by = identified_by
+        if identified_by == 'ep':
+            match = re.match(r'S(\d+)E(\d+)', ep_id)
+            episode.season = int(match.group(1))
+            episode.number = int(match.group(2))
+        elif identified_by == 'sequence':
+            episode.season = 0
+            episode.number = ep_id
+        series.episodes.append(episode)
+        # Need to flush to get an id on new Episode before assigning it as series begin
+        session.flush()
+    series.begin = episode
+
+
 def forget_series(name):
     """Remove a whole series `name` from database."""
     session = Session()
-    series = session.query(Series).filter(Series.name == name).first()
-    if series:
-        session.delete(series)
+    deleted = session.query(Series).filter(Series.name == name).delete()
+    if deleted:
         session.commit()
         log.debug('Removed series %s from database.', name)
     else:
@@ -576,14 +654,15 @@ class FilterSeriesBase(object):
                 },
                 # Strict naming
                 'exact': {'type': 'boolean'},
-                # Watched
-                'watched': {
-                    'type': ['string', 'object'],
-                    # SxxEyy form
-                    'pattern': '(?i)^s\d\de\d\d$',
-                    # dict form
-                    'properties': {'season': {'type': 'integer'}, 'episode': {'type': 'integer'}},
-                    'additionalProperties': False
+                # Begin takes an ep, sequence or date identifier
+                'begin': {
+                    'oneOf': [
+                        {'name': 'ep identifier', 'type': 'string', 'pattern': r'(?i)^S\d{2}E\d{2,3}$',
+                         'error_pattern': 'episode identifiers should be in the form `SxxEyy`'},
+                        {'name': 'date identifier', 'type': 'string', 'pattern': r'^\d{4}-\d{2}-\d{2}$',
+                         'error_pattern': 'date identifiers must be in the form `YYYY-MM-DD`'},
+                        {'name': 'sequence identifier', 'type': 'integer', 'minimum': 0}
+                    ]
                 },
                 'from_group': one_or_more({'type': 'string'}),
                 'parse_only': {'type': 'boolean'}
@@ -695,6 +774,19 @@ class FilterSeriesBase(object):
         return task.config['series']
 
 
+@event('manager.execute.started')
+def remove_old_tasks(manager):
+    """Unmark series from tasks which have been deleted"""
+    session = Session()
+    try:
+        deleted = (session.query(SeriesTask).filter(not_(SeriesTask.name.in_(manager.config['tasks']))).
+                   delete(synchronize_session=False))
+        if deleted:
+            session.commit()
+    finally:
+        session.close()
+
+
 class FilterSeries(SeriesDatabase, FilterSeriesBase):
     """
     Intelligent filter for tv-series.
@@ -702,15 +794,6 @@ class FilterSeries(SeriesDatabase, FilterSeriesBase):
     http://flexget.com/wiki/Plugins/series
     """
 
-    def __init__(self):
-        self.backlog = None
-
-    def on_process_start(self, task):
-        try:
-            self.backlog = get_plugin_by_name('backlog').instance
-        except DependencyError:
-            log.warning('Unable utilize backlog plugin, episodes may slip trough timeframe')
-
     @property
     def schema(self):
         return {
@@ -742,6 +825,15 @@ class FilterSeries(SeriesDatabase, FilterSeriesBase):
             }
         }
 
+    def __init__(self):
+        self.backlog = None
+
+    def on_process_start(self, task):
+        try:
+            self.backlog = get_plugin_by_name('backlog').instance
+        except DependencyError:
+            log.warning('Unable utilize backlog plugin, episodes may slip trough timeframe')
+
     def auto_exact(self, config):
         """Automatically enable exact naming option for series that look like a problem"""
 
@@ -792,14 +884,12 @@ class FilterSeries(SeriesDatabase, FilterSeriesBase):
                 continue
             # Make sure number shows (e.g. 24) are turned into strings
             series_name = unicode(series_name)
-            # Update database with capitalization from config
             db_series = task.session.query(Series).filter(Series.name == series_name).first()
-            if db_series:
-                db_series.name = series_name
-            else:
+            if not db_series:
                 log.debug('adding series %s into db', series_name)
                 db_series = Series()
                 db_series.name = series_name
+                db_series.identified_by = series_config.get('identified_by', 'auto')
                 task.session.add(db_series)
                 log.debug('-> added %s' % db_series)
             if not series_name in found_series:
@@ -933,9 +1023,11 @@ class FilterSeries(SeriesDatabase, FilterSeriesBase):
             log.debug('start with episodes: %s', [e['title'] for e in entries])
 
             # reject episodes that have been marked as watched in config file
-            if 'watched' in config:
-                log.debug('-' * 20 + ' watched -->')
-                if self.process_watched(config, ep, entries):
+            if ep.series.begin:
+                if ep < ep.series.begin:
+                    for entry in entries:
+                        entry.reject('Episode `%s` is before begin value of `%s`' %
+                                     (ep.identifier, ep.series.begin.identifier))
                     continue
 
             # skip special episodes if special handling has been turned off
@@ -1009,7 +1101,8 @@ class FilterSeries(SeriesDatabase, FilterSeriesBase):
                     log.debug('episode advancement disabled')
                 else:
                     log.debug('-' * 20 + ' episode advancement -->')
-                    if self.process_episode_advancement(ep, entries):
+                    # Grace is number of distinct eps in the task for this series + 2
+                    if self.process_episode_advancement(ep, entries, grace=len(series_entries)+2):
                         continue
 
             # quality
@@ -1032,7 +1125,7 @@ class FilterSeries(SeriesDatabase, FilterSeriesBase):
                     continue
 
             # Just pick the best ep if we get here
-            reason = reason or 'choosing best quality'
+            reason = reason or 'choosing best available quality'
             best.accept(reason)
 
     def process_propers(self, config, episode, entries):
@@ -1124,23 +1217,7 @@ class FilterSeries(SeriesDatabase, FilterSeriesBase):
             log.debug('no quality meets requirements')
         return result
 
-    def process_watched(self, config, episode, entries):
-        """
-        Rejects all episodes older than defined in watched.
-
-        :returns: True when rejected because of watched
-        """
-        from sys import maxint
-        w_season = config['watched'].get('season', -1)
-        w_episode = config['watched'].get('episode', maxint)
-        if episode.season < w_season or (episode.season == w_season and episode.number <= w_episode):
-            log.debug('%s episode %s is already watched, rejecting all occurrences',
-                      episode.series.name, episode.identifier)
-            for entry in entries:
-                entry.reject('watched')
-            return True
-
-    def process_episode_advancement(self, episode, entries):
+    def process_episode_advancement(self, episode, entries, grace):
         """Rejects all episodes that are too old or new (advancement), return True when this happens."""
 
         latest = self.get_latest_download(episode.series)
@@ -1148,15 +1225,15 @@ class FilterSeries(SeriesDatabase, FilterSeriesBase):
         log.debug('current: %s' % episode)
 
         if latest and latest.identified_by == episode.identified_by:
-            # allow few episodes "backwards" in case of missed eps
-            grace = len(entries) + 2
-            if ((episode.season < latest.season) or
-               (episode.season == latest.season and episode.number < (latest.number - grace))):
+            # Allow any previous episodes this season, or previous episodes within grace if sequence mode
+            if (episode.season < latest.season or
+                    (episode.identified_by == 'sequence' and episode.number < (latest.number - grace))):
                 log.debug('too old! rejecting all occurrences')
                 for entry in entries:
                     entry.reject('Too much in the past from latest downloaded episode %s' % latest.identifier)
                 return True
 
+            # Allow future episodes within grace, or first episode of next season
             if (episode.season > latest.season + 1 or (episode.season > latest.season and episode.number > 1) or
                (episode.season == latest.season and episode.number > (latest.number + grace))):
                 log.debug('too new! rejecting all occurrences')
@@ -1268,8 +1345,47 @@ class FilterSeries(SeriesDatabase, FilterSeriesBase):
                 log.debug('%s is not a series', entry['title'])
 
 
+class SeriesDBManager(FilterSeriesBase):
+    """Update in the database with series info from the config"""
+
+    @priority(0)
+    def on_task_start(self, task, config):
+        if not task.config_modified:
+            return
+        # Clear all series from this task
+        task.session.query(SeriesTask).filter(SeriesTask.name == task.name).delete()
+        if not task.config.get('series'):
+            return
+        config = self.prepare_config(task.config['series'])
+        for series_item in config:
+            series_name, series_config = series_item.items()[0]
+            # Make sure number shows (e.g. 24) are turned into strings
+            series_name = unicode(series_name)
+            db_series = task.session.query(Series).filter(Series.name == series_name).first()
+            if db_series:
+                # Update database with capitalization from config
+                db_series.name = series_name
+            else:
+                log.debug('adding series %s into db', series_name)
+                db_series = Series()
+                db_series.name = series_name
+                task.session.add(db_series)
+                log.debug('-> added %s' % db_series)
+            db_series.in_tasks.append(SeriesTask(task.name))
+            if series_config.get('identified_by', 'auto') != 'auto':
+                db_series.identified_by = series_config['identified_by']
+            # Set the begin episode
+            if series_config.get('begin'):
+                try:
+                    set_series_begin(db_series, series_config['begin'])
+                except ValueError as e:
+                    raise PluginError(e)
+
+
 # Register plugin
 register_plugin(FilterSeries, 'series')
+# This is a builtin so that it can update the database for tasks that may have had series plugin removed
+register_plugin(SeriesDBManager, 'series_db', builtin=True, api_ver=2)
 register_parser_option('--stop-waiting', action='store', dest='stop_waiting', default='',
                        metavar='NAME', help='Stop timeframe for a given series.')
 register_parser_option('--disable-advancement', action='store_true', dest='disable_advancement', default=False,
